import os
from io import BytesIO

import numpy as np
import pytest
from PIL import Image

from pyrogis.ingredients import Pierogi


@pytest.fixture
def array():
    return np.asarray([[[200, 200, 200], [30, 30, 30]],
                       [[130, 130, 130], [60, 60, 60]]]).astype(np.dtype('uint8'))


@pytest.fixture
def file(array: np.ndarray):
    image = Image.fromarray(array)
    memory = BytesIO()
    image.save(memory, format="PNG")

    return memory


@pytest.fixture
def path(array: np.ndarray, tmp_path):
    image = Image.fromarray(array)
    output_path = tmp_path / 'output.png'
    image.save(output_path)

    return output_path


def test_from_pil_image(file: BytesIO):
    """
    test that a Pierogi can be created from an Image
    """
    image = Image.open(file)
    pierogi = Pierogi.from_pil_image(image=image)

    assert np.all(pierogi.pixels[0, 0] == np.asarray([130, 130, 130]))
    assert np.all(pierogi.pixels[0, -1] == np.asarray([200, 200, 200]))
    assert np.all(pierogi.pixels[-1, 0] == np.asarray([60, 60, 60]))
    assert np.all(pierogi.pixels[-1, -1] == np.asarray([30, 30, 30]))


def test_from_path(path: str):
    """
    test that a Pierogi can be created from a file
    """

    pierogi = Pierogi.from_path(path=path)

    assert np.all(pierogi.pixels[0, 0] == np.asarray([130, 130, 130]))
    assert np.all(pierogi.pixels[0, -1] == np.asarray([200, 200, 200]))
    assert np.all(pierogi.pixels[-1, 0] == np.asarray([60, 60, 60]))
    assert np.all(pierogi.pixels[-1, -1] == np.asarray([30, 30, 30]))


def test_save(array: np.ndarray, tmp_path):
    """
    test resample can be used
    """

    pierogi = Pierogi(pixels=array)

    path = tmp_path / "output.png"
    pierogi.save(path)

    assert os.path.exists(path)


def test_resize(array: np.ndarray):
    """
    test resize method
    """

    pierogi = Pierogi(pixels=array)

    width = 100
    height = 200

    pierogi.resize(width=width, height=height)

    assert pierogi.width == width
    assert pierogi.height == height

    # nearest neighbor default
    assert np.all(pierogi.pixels[0, 0] == array[0, 0])
    assert np.all(pierogi.pixels[0, -1] == array[0, -1])
    assert np.all(pierogi.pixels[-1, 0] == array[-1, 0])
    assert np.all(pierogi.pixels[-1, -1] == array[-1, -1])


def test_resize_resample(array: np.ndarray):
    """
    test resample can be used
    """

    pierogi = Pierogi(pixels=array)

    width = 100
    height = 200

    pierogi.resize(width=width, height=height, resample=Image.BICUBIC)

    assert pierogi.width == width
    assert pierogi.height == height



"""
define an image wrapper ingredient
"""
from pathlib import Path
from typing import Callable, Union

import imageio
import numpy as np
from PIL import Image

from .ingredient import Ingredient


class Pierogi(Ingredient):
    """
    image container for iterative pixel manipulation

    create a Pierogi from:
    - pixel array
    - video file with a frame index
    - image file
    - PIL image

    unless pixels are provided explicitly ( Pierogi(pixels=pixels) ),
    the pixels member is a property that is lazy loaded
    (usually from a file) and cached
    """

    RESAMPLE = Image.NEAREST
    """default resize algorithm is nearest neighbor"""

    _pixels: np.ndarray = None
    """underlying numpy pixels array"""

    @property
    def image(self) -> Image.Image:
        """
        turn the numpy array into a PIL Image
        """
        image = Image.fromarray(np.rot90(self.pixels), 'RGB')
        return image

    @property
    def width(self) -> int:
        """
        1st dimension of the underlying pixel array
        """
        return self.pixels.shape[0]

    @property
    def height(self) -> int:
        """
        2nd dimension of the underlying pixel array
        """
        return self.pixels.shape[1]

    def prep(
            self,
            pixels: np.ndarray = None,
            loader: Callable[[], np.ndarray] = None,
            **kwargs
    ) -> None:
        """
        provide the source image in a number of ways

        :param pixels: numpy array
        :param loader: function that produces a pixels array
        """

        if pixels is not None:
            self._pixels = pixels
            self._loader = lambda: pixels

        elif loader is not None:
            self._loader = loader

        else:
            raise Exception("one of pixels or loader must be provided")

    @classmethod
    def from_path(cls, path: str, frame_index: int = 0) -> 'Pierogi':
        """
        :param path: file path to load from
        :param frame_index: if path is a multiframe format (video),
        use this specified frame
        """

        def loader():
            reader = imageio.get_reader(path)
            reader.set_image_index(frame_index)
            return np.rot90(np.array(reader.get_next_data())[:, :, :3], axes=(1, 0))

        return cls(loader=loader)

    @classmethod
    def from_shape(cls, shape: tuple) -> 'Pierogi':
        """
        :param shape: (width, height) to make default pixels array
        """

        def loader():
            return np.full((*shape, 3), cls._default_pixel)

        return cls(loader=loader)

    @classmethod
    def from_pil_image(cls, image: Image.Image):
        """
        :param image: PIL Image that has already been loaded
        """

        def loader():
            return np.rot90(np.array(image.convert('RGB')), axes=(1, 0))

        return cls(loader=loader)

    @property
    def pixels(self) -> np.ndarray:
        if self._pixels is None:
            self.load()

        return self._pixels

    def load(self) -> None:
        """
        use the loader return the contained pixels one time
        """
        self._pixels = self._loader()

    def cook(self, pixels: np.ndarray) -> np.ndarray:
        """
        performs actions on a pixel array and returns a cooked array
        """
        return self.pixels

    def show(self) -> None:
        """
        open an image viewer to display the array
        """
        self.image.show()

    def save(self, path: Union[str, Path], optimize: bool = False) -> None:
        """
        save the image to the given path
        """

        output_filename = path

        self.image.save(output_filename, optimize=optimize)

    def resize(self, width: int, height: int, resample: int = RESAMPLE):
        """
        resize pixels to new width and height

        :param width: width to resize to

        :param height: height to resize to

        :param resample: resample method to use in Image.resize.
        PIL documentation:

        >    "An optional resampling filter.
        >    This can be one of PIL.Image.NEAREST (use nearest neighbour),
        >    PIL.Image.BILINEAR (linear interpolation),
        >    PIL.Image.BICUBIC (cubic spline interpolation),
        >    or PIL.Image.LANCZOS (a high-quality downsampling filter).
        >    If omitted, or if the image has mode “1” or “P”, it is set PIL.Image.NEAREST."
        """

        if height != self.height and width != self.width:
            self._pixels = np.array(
                Image.fromarray(
                    self.pixels
                ).resize(
                    (height, width), resample
                )
            )



import numpy as np

from pyrogis.ingredients import (
    Dish, Ingredient, Pierogi,
    Quantize, Recipe, Sort, Threshold
)

pierogi = Pierogi.from_path('gnome.jpg')


def threshold_example():
    threshold = Threshold(upper_threshold=100)

    # pass in lists to be mixed
    threshold_recipe = Recipe(ingredients=[threshold])
    # recipe = Recipe(recipe)

    threshold_dish = Dish(
        pierogi=pierogi,
        recipe=threshold_recipe
    )

    cooked_dish = threshold_dish.serve()

    cooked_dish.pierogi.show()


def swap_example():
    threshold = Threshold(target=pierogi, upper_threshold=200)

    swap_pixels = np.random.randint(0, 1, (10, 10, 3))
    swap = Ingredient(swap_pixels)
    swap.season(threshold)

    swap_recipe = Recipe(ingredients=[swap])

    swap_dish = Dish(
        pierogi=pierogi,
        recipe=swap_recipe
    )

    cooked_dish = swap_dish.serve()

    cooked_dish.pierogi.show()


def quantize_example():
    palette = [
        [100, 210, 69],
        [45, 23, 180],
        [62, 31, 70],
        [10, 210, 240],
        [45, 10, 244],
        [38, 31, 10],
        [255, 255, 255],
        [99, 94, 124]
    ]
    quantize = Quantize(palette=palette)

    # pass in lists to be mixed
    quantize_recipe = Recipe(ingredients=[quantize])

    quantize_dish = Dish(
        pierogi=pierogi,
        recipe=quantize_recipe
    )

    cooked_dish = quantize_dish.serve()

    cooked_dish.pierogi.show()


def sort_example():
    # threshold is a seasoning, which means it can be
    # used to add a mask to another Ingredient
    threshold = Threshold(target=pierogi, upper_threshold=80)

    sort = Sort()
    # apply a threshold mask to the sort
    sort.season(threshold)

    sort_recipe = Recipe(ingredients=[pierogi, sort])

    sort_dish = Dish(
        pierogi=pierogi,
        recipe=sort_recipe
    )

    cooked_dish = sort_dish.serve()

    cooked_dish.pierogi.show()


if __name__ == '__main__':
    sort_example()



"""
these special ingredients are good for applying their cook method
to create a B/W mask for other ingredients to use internally
"""

from .seasoning import Seasoning
from .threshold import Threshold
from .rectangle import Rectangle



import os
from typing import List

import pytest

from pyrogis.kitchen import Kitchen, Server, Ticket
from pyrogis.kitchen.menu import ResizeFilling
from pyrogis.kitchen.order import Order


@pytest.fixture
def server() -> Server:
    return Server()


@pytest.fixture
def parsed_vars() -> dict:
    parsed_vars = {
        'filling': 'resize',
        'generate_ticket': ResizeFilling.generate_ticket
    }
    return parsed_vars


@pytest.fixture
def image_order(order_name, image_path) -> Order:
    return Order(order_name, image_path)


@pytest.fixture
def animation_order(order_name, animation_path) -> Order:
    return Order(order_name, animation_path)


@pytest.fixture
def dir_order(order_name, dir_path) -> Order:
    return Order(order_name, dir_path)


@pytest.fixture
def recipe_path(tmp_path):
    recipe_text = """
    resize -s .5;
    sort;
    quantize;
    resize -s 2;
    """
    recipe_path = tmp_path / 'recipe.txt'
    with open(recipe_path, 'w') as f:
        f.write(recipe_text)

    return str(recipe_path)


def test_write_tickets_image(server: Server, image_order: Order, parsed_vars):
    server._write_tickets(image_order, parsed_vars)

    assert len(image_order.tickets) == 1


def test_write_tickets_animation(server, animation_order: Order, parsed_vars):
    server._write_tickets(animation_order, parsed_vars)

    assert len(animation_order.tickets) > 1


def test_check_cooked(server: Server, image_order: Order, image_path: str):
    image_order.tickets = [Ticket(output_path=image_path)]
    server._check_order(order=image_order)


# take_order

def run_take_order(
        server: Server, kitchen: Kitchen, args: List[str]
) -> Order:
    """handle take order call and successful output checking"""
    order = server.take_order(args, kitchen)

    for output_path in order.ticket_output_paths:
        assert os.path.isfile(output_path)

    return order


def test_take_order_sort(server, kitchen, image_path):
    """test sort order"""
    args = ["sort", image_path]

    run_take_order(server, kitchen, args)


def test_take_order_sort_options(server, kitchen, image_path):
    """test sort order with options"""
    args = ["sort", image_path, "-u", "120", "-l", "20", "-t", "2", "--ccw"]

    run_take_order(server, kitchen, args)


def test_take_order_quantize(server, kitchen, image_path):
    """test quantize order"""
    args = ["quantize", image_path]
    run_take_order(server, kitchen, args)


def test_take_order_quantize_options(server, kitchen, image_path):
    """test quantize order with options"""
    args = [
        "quantize", image_path,
        "-c", "012312", "043251",
        "-n", "4",
        "--iterations", "2",
        "--repeats", "2",
        "--initial-temp", ".8",
        "--final-temp", "0.1",
        "--dithering-level", "0.5",
    ]

    run_take_order(server, kitchen, args)


def test_take_order_threshold(server, kitchen, image_path):
    args = ["threshold", image_path]

    run_take_order(server, kitchen, args)


def test_take_order_threshold_options(server, kitchen, image_path):
    """test threshold order with options"""
    args = [
        "threshold", image_path,
        "-u", "200",
        "-l", "20",
        "--include", "abaabb",
        "--exclude", "333433"
    ]

    run_take_order(server, kitchen, args)


def test_take_order_resize(server, kitchen, image_path):
    """test resize order"""
    args = ["resize", image_path]

    order = run_take_order(server, kitchen, args)

    assert len(order.tickets) == 1
    print(order.tickets[0].output_path)


def test_take_order_resize_options(server, kitchen, image_path):
    """test resize order with options"""
    args = [
        "resize", image_path,
        "--width", "200",
        "--height", "300",
        "-s", "2",
        "--resample-filter", "bicubic",
        "--presave",
        "--async",
        "--processes", "4"
    ]

    order = run_take_order(server, kitchen, args)

    assert kitchen.processes == 4
    assert order.cook_async
    assert order.presave


def test_take_order_presave_async_processes(server, kitchen, image_path):
    """test queue/async options"""
    args = [
        "resize", image_path,
        "--presave",
        "--async",
        "--processes", "4"
    ]

    run_take_order(server, kitchen, args)


def test_take_order_resume(server, kitchen, image_path):
    """test resume flag"""
    args = [
        "resize", image_path,
        "--resume"
    ]

    run_take_order(server, kitchen, args)


def test_take_order_custom(server, kitchen, image_path):
    """test custom with a txt file as a recipe"""
    args = ["custom", image_path, "sort; quantize"]

    run_take_order(server, kitchen, args)


def test_take_order_custom_txt(server, kitchen, image_path, recipe_path):
    """test custom with a txt file as a recipe"""
    args = ["custom", image_path, recipe_path]

    run_take_order(server, kitchen, args)


def test_take_order_crop(server, kitchen, image_path, recipe_path):
    """test custom with a txt file as a recipe"""
    args = [
        "crop", image_path,
        "--width", "1",
        "--height", "1",
    ]

    run_take_order(server, kitchen, args)


def test_take_order_crop_options(server, kitchen, image_path, recipe_path):
    """test custom with a txt file as a recipe"""
    args = [
        "crop", image_path,
        "--origin", "ne",
        "--width", "1",
        "--height", "1",
    ]

    run_take_order(server, kitchen, args)


def test_take_order_crop_options_float(server, kitchen, image_path, recipe_path):
    """test custom with a txt file as a recipe"""
    args = [
        "crop", image_path,
        "--origin", "ne",
        "--width", "1",
        "--height", "1",
        "-x", "-0.5",
        "--aspect", "0.5",
    ]

    run_take_order(server, kitchen, args)


# non filling-related take_order tests
def test_take_order_togo(server, kitchen, dir_path):
    """test togo"""
    args = ["togo", dir_path]

    run_take_order(server, kitchen, args)


def test_take_order_togo_options(server, kitchen, dir_path, mp4_output_path):
    """test togo with options"""
    args = [
        "togo", dir_path,
        "--fps", "25",
        "--frame-duration", "20",
        "--no-optimize",
        "--output", mp4_output_path,
        "--order-name", "octo"
    ]

    run_take_order(server, kitchen, args)


def test_take_order_image_with_output(server, kitchen, image_path, png_output_path):
    """test image output path given"""
    args = ["resize", image_path, "--output", png_output_path]

    run_take_order(server, kitchen, args)


def test_take_order_animation(server, kitchen, animation_path):
    """test making an animation order"""
    args = ["resize", animation_path]

    run_take_order(server, kitchen, args)



import os
import sys

from importlib_metadata import version

sys.path.insert(0, os.path.abspath('../src'))

# Project --------------------------------------------------------------

project = "pierogis"
copyright = "2021 pierogis"
author = "pierogis-live"
version = version('pyrogis')

# General --------------------------------------------------------------

master_doc = "index"
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinxcontrib.apidoc',
    'sphinx_issues',
    'sphinx.ext.viewcode'
]
intersphinx_mapping = {
    "rich": ("https://rich.readthedocs.io/en/stable/", None),
    "imageio": ("https://imageio.readthedocs.io/en/stable/", None),
    "Pillow": ("https://pillow.readthedocs.io/en/stable/", None),
    'python': ('https://docs.python.org/3', None),
}
issues_github_path = "pierogis/pierogis"
apidoc_module_dir = '../src/pyrogis'
apidoc_output_dir = 'source'
apidoc_separate_modules = True

# HTML -----------------------------------------------------------------

html_theme = "alabaster"
html_sidebars = {
    '**': [
        'about.html',
        'navigation.html',
        'relations.html',
        'searchbox.html',
    ]
}
html_title = f"pierogis docs ({version})"
html_show_sourcelink = True
html_theme_options = {
    "description": "image and animation processing framework",
    "github_user": "pierogis",
    "github_repo": "pierogis",
    "fixed_sidebar": True,
    'github_type': ''
}



from abc import abstractmethod
from typing import Dict, List, Protocol

from .menu import Filling
from .ticket import Ticket, PierogiDesc, IngredientDesc
from ..ingredients import (
    Ingredient, Dish, Pierogi, Recipe
)


class Cooker(Protocol):
    @abstractmethod
    def assemble_ticket(self, ticket: Ticket, menu: Dict) -> Dish:
        pass

    @abstractmethod
    def cook_dish(
            self,
            dish: Dish
    ) -> Dish:
        pass


class Chef(Cooker):
    """gets dishes from tickets and cooks"""

    @classmethod
    def create_pierogi_objects(
            cls,
            pierogi_descs: Dict[str, PierogiDesc],
            files: Dict[str, str]
    ) -> Dict[str, Pierogi]:
        """
        exchange a set of PierogiDesc keyed by a name
        with Pierogi loaded from file links
        """
        pierogi_objects = {}

        for pierogi_key, pierogi_desc in pierogi_descs.items():
            pierogi_obj = pierogi_desc.create(files)

            pierogi_objects[pierogi_key] = pierogi_obj

        return pierogi_objects

    @classmethod
    def create_ingredient_objects(
            cls,
            ingredient_descs: Dict[str, IngredientDesc],
            pierogis: Dict[str, Pierogi],
            menu: Dict[str, Filling]
    ) -> Dict[str, Ingredient]:
        """
        create a dict of uuid->Ingredient
        from an ingredient description and a lookup for file paths

        :param ingredient_descs: map of uuid keys to inner dict
        with type, args, and kwargs keys
        :param pierogis: map of uuid keys created Pierogi
        """
        ingredients: Dict[str, Ingredient] = {}

        for ingredient_name, ingredient_desc in ingredient_descs.items():
            # if path is one of the kwargs
            # look it up in the linking paths dictionary
            pierogi_name = ingredient_desc.kwargs.get('pierogi')
            if pierogi_name is not None:
                pierogi = pierogis[pierogi_name]
                ingredient_desc.kwargs['pierogi'] = pierogi

            ingredient = cls.get_ingredient(
                ingredients, ingredient_descs, ingredient_name, menu
            )

            ingredients[ingredient_name] = ingredient

        return ingredients

    @classmethod
    def get_ingredient(
            cls,
            ingredients: dict,
            ingredient_descs: Dict[str, IngredientDesc],
            ingredient_name: str,
            menu: dict
    ) -> Ingredient:
        """
        look to see if an ingredient object has already been created
        otherwise create it and swap it in the ingredients dictionary
        """
        ingredient = ingredients.get(ingredient_name)

        if ingredient is None:
            ingredient_desc = ingredient_descs[ingredient_name]

            # search kwargs for keys that are type names
            # swap the value of that kwarg (reference key)
            # for the created ingredient obj
            for ingredient_type_name in menu.keys():
                ingredient_name = ingredient_desc.kwargs.get(
                    ingredient_type_name
                )

                if ingredient_name is not None:
                    ingredient = cls.get_ingredient(
                        ingredients, ingredient_descs, ingredient_name, menu
                    )
                    ingredient_desc.kwargs[ingredient_type_name] = ingredient

            # now create an ingredient as specified in the description
            ingredient = ingredient_desc.create(menu)

        return ingredient

    @classmethod
    def apply_seasonings(cls, ingredients, seasoning_links) -> None:
        for seasoning, recipient in seasoning_links.items():
            seasoning = ingredients[seasoning]
            recipient = ingredients[recipient]

            recipient.season(seasoning)

    @classmethod
    def create_recipe_object(
            cls,
            ingredients: Dict[str, Ingredient],
            recipe_order: List[str]
    ) -> Recipe:
        """
        create a recipe from:
        ingredients map
        seasoning name map
        order of ingredients

        :param ingredients: map of uuid key to ingredient object value
        :param recipe_order: order of ingredients by uuid
        """

        recipe = Recipe(ingredients=[])
        # loop through the ingredient keys specified by the recipe
        for ingredient_key in recipe_order:
            # there might be a bug here with the ingredient being out
            # of sync with another reference to that ingredient
            ingredient = ingredients[ingredient_key]

            # add this created ingredient to the dish recipe for return
            recipe.add(ingredient)

        return recipe

    @classmethod
    def assemble_ticket(cls, ticket: Ticket, menu: Dict[str, Filling]) -> Dish:
        """
        cook a dish from a series of descriptive dicts
        """
        pierogi_descs = ticket.pierogis
        ingredient_descs = ticket.ingredients
        files = ticket.files
        recipe = ticket.recipe
        base = ticket.base
        seasoning_links = ticket.seasoning_links

        pierogis = cls.create_pierogi_objects(
            pierogi_descs,
            files
        )

        ingredients = cls.create_ingredient_objects(
            ingredient_descs,
            pierogis,
            menu
        )

        cls.apply_seasonings(
            ingredients,
            seasoning_links
        )

        recipe_object = cls.create_recipe_object(
            ingredients,
            recipe
        )

        return Dish(
            pierogi=pierogis[base],
            recipe=recipe_object
        )

    @classmethod
    def cook_dish(
            cls,
            dish: Dish
    ) -> Dish:
        """

        """
        return dish.serve()



from .filling import Filling
from ...ingredients import Crop
from ...ingredients.seasonings.rectangle import Direction, Rectangle


class CropFilling(Filling):
    type_name = 'crop'
    type = Crop

    @classmethod
    def add_parser_arguments(cls, parser):
        """parse for crop instructions"""
        parser.add_argument(
            '--origin',
            default=Rectangle.ORIGIN,
            type=Direction,
            choices=list(Direction),
            help="origin location; sw -> (0,0); ne -> (input width,input height)... etc"
        )
        parser.add_argument(
            '--height',
            type=int,
            help="height of crop selection"
        )
        parser.add_argument(
            '--width',
            type=int,
            help="width of crop selection"
        )
        parser.add_argument(
            '--aspect',
            type=float,
            help="aspect ratio; ignored if height and width provided"
        )

        def int_or_float(value: str):
            f = float(value)
            if f.is_integer():
                return int(f)
            else:
                return f

        parser.add_argument(
            '-x',
            default=Crop.X,
            type=int_or_float,
            help="x offset from the origin; can be negative;"
                 "if a float, corresponds to that percentage of the input image width"
        )
        parser.add_argument(
            '-y',
            default=Crop.Y,
            type=int_or_float,
            help="y offset from the origin; can be negative;"
                 "if a float, corresponds to that percentage of the input image height"
        )



import multiprocessing as mp
import sys
from typing import Callable

from .kitchen import Chef, Server, Kitchen
from .restaurant import Restaurant


def run(args=None, report_callback: Callable = None):
    if args is None:
        args = sys.argv[1:]

    server = Server(report_callback)

    mp.set_start_method('spawn', force=True)
    kitchen = Kitchen(Chef())

    server.take_order(
        args,
        kitchen
    )


def main():
    """cli program"""

    restaurant = Restaurant()
    sys.exit(restaurant.open(run))


if __name__ == '__main__':
    main()



import numpy as np

from .ingredient import Ingredient
from .pierogi import Pierogi
from .recipe import Recipe


class Dish(Ingredient):
    """
    cook a pierogi with a recipe
    """

    _pierogi: Pierogi = None

    def prep(
            self,
            pierogi: Pierogi,
            recipe: Recipe = None,
            **kwargs
    ):
        """
        set the recipe to cook for this dish

        :param recipe: something callable that returns a
        cook(pixels) method.
        Any Ingredient (including recipe) is an example of this

        :param pierogis: a list of Pierogi to cook
        """
        self.pierogi = pierogi

        if recipe is None:
            recipe = Recipe()
        self.recipe = recipe

    def cook(self, pixels: np.ndarray) -> np.ndarray:
        return self.recipe.cook(self.pierogi.pixels)

    def serve(self) -> 'Dish':
        """
        cook the recipe and set the output to this object's pixel array
        """

        # cook with these pixels as first input
        cooked_pixels = self.recipe.cook(self.pierogi.pixels)
        # ensure that the cooked pixels do not overflow 0-255
        clipped_pixels = np.clip(cooked_pixels, 0, 255)
        # # set the objects own pixels to the result of cooking
        cooked_pierogi = Pierogi(pixels=clipped_pixels)

        return Dish(pierogi=cooked_pierogi)



import numpy as np
import pytest

from pyrogis.ingredients.crop import Crop


@pytest.fixture
def array():
    return np.asarray([[[200, 200, 200], [30, 30, 30], [45, 90, 180]],
                       [[130, 130, 130], [60, 60, 60], [10, 90, 50]],
                       [[20, 40, 100], [10, 20, 200], [220, 25, 240]]]).astype(np.dtype('uint8'))


def test_crop_width(array):
    """
    provide only width
    """
    width = 1

    crop = Crop(width=width)
    cooked_array = crop.cook(array)

    assert cooked_array.shape[0] == width
    assert cooked_array.shape[1] == array.shape[1]


def test_crop_height(array):
    """
    provide only height
    """
    height = 2

    crop = Crop(height=height)
    cooked_array = crop.cook(array)

    assert cooked_array.shape[0] == array.shape[0]
    assert cooked_array.shape[1] == height


def test_crop_aspect_1(array):
    """
    provide only aspect with height limiting
    """
    aspect = 2 / 3

    crop = Crop(aspect=aspect)
    cooked_array = crop.cook(array)

    assert cooked_array.shape[0] == int(aspect * array.shape[0])
    assert cooked_array.shape[1] == array.shape[1]


def test_crop_aspect_2(array):
    """
    provide only aspect with width limiting
    """
    aspect = 3 / 2

    crop = Crop(aspect=aspect)
    cooked_array = crop.cook(array)

    assert cooked_array.shape[0] == array.shape[0]
    assert cooked_array.shape[1] == int(array.shape[1] / aspect)


def test_crop_width_height(array):
    """
    provide height and width
    """
    width = 2
    height = 1

    crop = Crop(width=width, height=height)
    cooked_array = crop.cook(array)

    assert cooked_array.shape[0] == width
    assert cooked_array.shape[1] == height


def test_crop_width_aspect(array):
    """
    provide width and aspect
    """
    width = 1
    aspect = 1 / 2

    crop = Crop(width=width, aspect=aspect)
    cooked_array = crop.cook(array)

    assert cooked_array.shape[0] == width
    assert cooked_array.shape[1] == width / aspect


def test_crop_height_aspect(array):
    """
    provide height and aspect
    """
    height = 1
    aspect = 2

    crop = Crop(aspect=aspect, height=height)
    cooked_array = crop.cook(array)

    assert cooked_array.shape[0] == height * aspect
    assert cooked_array.shape[1] == height


def test_crop_width_height_aspect(array):
    """
    provide width height and aspect

    just ignores aspect
    maybe it should fail
    """
    width = 2
    height = 1
    aspect = 2
    aspect = array.shape[0] / array.shape[1]

    crop = Crop(width=width, height=height, aspect=aspect)
    cooked_array = crop.cook(array)

    assert cooked_array.shape[0] == width
    assert cooked_array.shape[1] == height


def test_crop_x_y(array):
    """
    provide origin
    """
    x = 1
    y = 1

    crop = Crop(x=x, y=y)
    cooked_array = crop.cook(array)

    assert cooked_array.shape[0] == array.shape[0] - x
    assert cooked_array.shape[1] == array.shape[1] - y


def test_crop_x_y_height(array):
    """
    provide origin
    """
    x = 1
    y = 1
    height = 1

    crop = Crop(x=x, y=y, height=height)
    cooked_array = crop.cook(array)

    assert cooked_array.shape[0] == array.shape[0] - x
    assert cooked_array.shape[1] == height



import enum
from collections import namedtuple

Coordinate = namedtuple('Coordinate', ['x', 'y'])


class Direction(enum.Enum):
    NE = 'ne'
    NW = 'nw'
    SE = 'se'
    SW = 'sw'
    N = 'n'
    E = 'e'
    S = 's'
    W = 'w'
    C = 'c'

    def __str__(self):
        return self.value



from .filling import Filling
from ...ingredients import Threshold


class ThresholdFilling(Filling):
    type_name = 'threshold'
    type = Threshold

    @classmethod
    def add_parser_arguments(cls, parser):
        """
        parse for thresholds
        """
        parser.add_argument(
            '-l', '--lower-threshold',
            default=Threshold.LOWER_THRESHOLD, type=int,
            help="Pixels with lightness below"
                 "this threshold are included"
        )
        parser.add_argument(
            '-u', '--upper-threshold',
            default=Threshold.UPPER_THRESHOLD, type=int,
            help="Pixels with lightness above"
                 "this threshold are included"
        )
        parser.add_argument(
            '--include',
            dest='include',
            default="ffffff", type=str,
            help="Hex color for included pixels"
        )
        parser.add_argument(
            '--exclude',
            dest='exclude',
            default="000000", type=str,
            help="Hex color for excluded pixels"
        )



import numpy as np

from .ingredient import Ingredient
from .seasonings.cartography import Direction
from .seasonings.rectangle import Rectangle


class Crop(Ingredient):
    """
    crop starting from an origin and selecting an area both defined as a compass direction
    """
    X = 0
    Y = 0

    x: int
    y: int
    height: int
    width: int
    aspect: float

    def prep(
            self,
            x: int = X,
            y: int = Y,
            height: int = None,
            width: int = None,
            aspect: float = None,
            origin: Direction = Rectangle.ORIGIN,
            **kwargs
    ) -> None:
        """
        :param x: 0 = left
        :param y: 0 = bottom
        :param width: width in pixels
        :param height: height in pixels
        :param aspect: aspect ratio to fill missing height or width (width/height, 1 means square)
        :param origin: location of origin for direction of crop select
        """
        self.x = x
        self.y = y
        self.height = height
        self.width = width
        self.aspect = aspect
        self.origin = origin

    def cook(self, pixels: np.ndarray) -> np.ndarray:
        width = pixels.shape[0]
        height = pixels.shape[1]

        rectangle = Rectangle(
            width=self.width, height=self.height,
            x=self.x, y=self.y,
            aspect=self.aspect, origin=self.origin
        )

        bottom_left, top_right = rectangle.get_corner_coordinates(width, height)

        cooked_pixels = pixels[
                        bottom_left.x: top_right.x,
                        bottom_left.y: top_right.y
                        ]

        return cooked_pixels



"""
parsing
"""
import argparse
import os
import time
from abc import abstractmethod
from threading import Thread
from typing import List, Callable, Protocol

import imageio

from .kitchen import Kitchen
from .order import Order
from .ticket import Ticket


class OrderTaker(Protocol):
    """any object that implements take_order and check_order is an OrderTaker"""

    @abstractmethod
    def take_order(
            self, args: List[str], kitchen: Kitchen
    ) -> Order:
        pass


class Server(OrderTaker):
    """handles tasks related to taking input and preparing it for Kitchen"""

    def __init__(
            self,
            report_callback: Callable = None,
    ):
        self._report_callback = report_callback

    def _create_parser(self, menu):
        # create top level parser
        parser = argparse.ArgumentParser(
            description='** image processing pipelines **'
        )
        subparsers = parser.add_subparsers(
            dest='filling', required=True
        )

        # create parent parser to pass down arguments only
        base_parser = argparse.ArgumentParser()
        base_parser.add_argument(
            'path',
            default='./',
            help="path to file or directory to use as input"
        )
        base_parser.add_argument(
            '--order-name',
            help="""
            name of order for filename prefixes; used to identify files for output
            """
        )

        subparsers.add_parser('togo', parents=[base_parser], add_help=False)

        for command, menu_item in menu.items():
            # inherit the parent class arguments
            # and arguments specific to a subcommand
            subparsers.add_parser(
                command,
                parents=[base_parser, menu_item.get_parser()],
                add_help=False
            )

        return parser

    def _create_togo_parser(self):
        # add parser options for outputting as animation (gif)
        togo_parser = argparse.ArgumentParser(add_help=False)
        togo_parser.add_argument(
            '-o', '--output',
            dest='output_path',
            help="path to save resulting image"
        )
        togo_parser.add_argument(
            '--frame-duration',
            type=int,
            help="frame duration in ms"
        )
        togo_parser.add_argument(
            '--fps',
            type=int,
            help="frames per second"
        )
        togo_parser.add_argument(
            '--no-optimize',
            dest='optimize',
            default=True,
            action='store_false',
            help="duration in ms"
        )

        return togo_parser

    def _write_tickets(
            self, order: Order, parsed_vars
    ) -> None:
        """create tickets from a list of pierogis and parsed vars"""
        generate_ticket = parsed_vars.pop('generate_ticket')
        parsed_vars.pop('filling')

        input_path = order.input_path
        order_name = order.order_name

        if os.path.isfile(input_path):
            reader = imageio.get_reader(input_path)
            if hasattr(reader, 'count_frames'):
                frames = reader.count_frames()
            else:
                frames = len(reader)

            for frame_index in range(frames):
                ticket = Ticket()
                ticket = generate_ticket(ticket, input_path, frame_index, **parsed_vars.copy())

                order.add_ticket(ticket)

        elif os.path.isdir(input_path):
            # debug here
            filenames = [
                filename
                for filename
                in os.listdir(input_path)
                if filename != ".DS_Store"
            ]
            for filename in sorted(filenames):
                if order_name is None or filename.startswith(order_name):
                    ticket_input_path = os.path.join(
                        input_path,
                        filename
                    )

                    ticket = Ticket()
                    ticket = generate_ticket(ticket, ticket_input_path, 0, **parsed_vars.copy())

                    order.add_ticket(ticket)

        else:
            raise FileNotFoundError(input_path)

    def _handle_togo(self, order: Order):
        self._report_status(order, status='boxing')

        if os.path.isdir(order.input_path):
            # debug here
            filenames = [
                filename
                for filename
                in os.listdir(order.input_path)
                if filename != ".DS_Store"
            ]
            for filename in sorted(filenames):
                if order.order_name is None or filename.startswith(order.order_name):
                    frame_path = os.path.join(order.input_path, filename)
                    ticket = Ticket(output_path=frame_path)
                    order.add_ticket(ticket)
        else:
            for i in range(imageio.get_reader(order.input_path).count_frames()):
                ticket = Ticket()
                order.add_ticket(ticket)

        frames = len(order.tickets)
        self._report_status(order, total=frames)

    def _handle_filling(self, order: Order, parsed_vars: dict, kitchen: Kitchen):
        self._report_status(order, status='writing')

        presave = parsed_vars.pop('presave')
        cook_async = parsed_vars.pop('async')
        processes = parsed_vars.pop('processes')
        resume = parsed_vars.pop('resume')

        order.presave = presave
        order.cook_async = cook_async
        order.processes = processes
        order.resume = resume

        # order has tickets attached (for frames)
        self._write_tickets(order, parsed_vars)

        frames = len(order.tickets)
        self._report_status(order, total=frames)

        # the Server splits off part of their consciousness into a new thread
        check_thread = Thread(target=self._check_order, args=[order])
        check_thread.daemon = True

        # don't start the checker thread until the kitchen is ready
        def start_callback():
            check_thread.start()

        self._report_status(order, status='cooking')

        kitchen.queue_order(order, start_callback, self._report_status)
        check_thread.join()

        self._report_status(order, status='boxing')

    def take_order(
            self, args: List[str], kitchen: Kitchen
    ) -> Order:
        """
        take a list of args and turn it into an Order for the Kitchen
        """

        togo_parser = self._create_togo_parser()
        parsed, unknown = togo_parser.parse_known_args(args)
        parsed_togo_vars = vars(parsed)

        parser = self._create_parser(kitchen.menu)
        # parse the input args with the applicable arguments attached
        parsed = parser.parse_args(unknown)
        parsed_vars = vars(parsed)

        # need the path to use as input for some recipes
        # like opening files for ingredients
        input_path = os.path.expanduser(os.path.abspath(parsed_vars.pop('path')))
        order_name = parsed_vars.pop('order_name')

        output_path = parsed_togo_vars.pop('output_path')
        fps = parsed_togo_vars.pop('fps')
        optimize = parsed_togo_vars.pop('optimize')
        frame_duration = parsed_togo_vars.pop('frame_duration')

        order = Order(
            order_name, input_path,
            output_path=output_path,
            fps=fps,
            duration=frame_duration,
            optimize=optimize
        )

        if order.fps is None:
            if os.path.isfile(input_path):
                order.fps = imageio.get_reader(input_path).get_meta_data().get('fps')

        # if the order is just togo, don't need the kitchen
        if parsed_vars['filling'] == 'togo':
            self._handle_togo(order)
        else:
            self._handle_filling(order, parsed_vars, kitchen)

        kitchen.plate(
            order=order
        )

        self._report_status(order, status='done')

        return order

    @staticmethod
    def _count_cooked_tickets(order: Order) -> int:
        cooked_tickets = 0
        for filename in order.ticket_output_paths:
            if os.path.exists(os.path.join(filename)):
                cooked_tickets += 1

        return cooked_tickets

    def _check_order(
            self,
            order: Order
    ) -> int:
        """check and report how many tickets from an Order have been cooked"""
        total = len(order.tickets)

        wait_time = .1

        failed_output_paths = []

        while True:
            completed = self._count_cooked_tickets(order)

            while not order.failures.empty():
                exception, ticket = order.failures.get()
                failed_output_paths.append(ticket.output_path)

            self._report_status(
                order,
                completed=completed,
                branches=failed_output_paths
            )

            if completed + len(failed_output_paths) == total:
                break
            else:
                time.sleep(wait_time)

        return completed

    def _report_status(self, order, **kwargs) -> None:
        """
        currently pushes the status to a configured collback

        this could also log to a file
        """

        if self._report_callback is not None:
            self._report_callback(order, **kwargs)



import numpy as np
import pytest

from pyrogis.ingredients.seasonings import Rectangle
from pyrogis.ingredients.seasonings.rectangle import Direction


@pytest.fixture
def array():
    return np.zeros((500, 500, 3))


def test_get_corner_coordinates(array: np.ndarray):
    rectangle = Rectangle()
    bottom_left, top_right = rectangle.get_corner_coordinates(array.shape[0], array.shape[1])

    assert bottom_left.x == 0
    assert bottom_left.y == 0
    assert top_right.x == 500
    assert top_right.y == 500


def test_get_corner_coordinates_width_height_sw(array: np.ndarray):
    origin = Direction.SW
    width = 100
    height = 100

    rectangle = Rectangle(origin=origin, width=width, height=height)
    bottom_left, top_right = rectangle.get_corner_coordinates(array.shape[0], array.shape[1])

    assert bottom_left.x == 0
    assert bottom_left.y == 0
    assert top_right.x == 100
    assert top_right.y == 100


def test_get_corner_coordinates_width_height_se(array: np.ndarray):
    origin = Direction.SE
    width = 100
    height = 100

    rectangle = Rectangle(origin=origin, width=width, height=height)
    bottom_left, top_right = rectangle.get_corner_coordinates(array.shape[0], array.shape[1])

    assert bottom_left.x == 400
    assert bottom_left.y == 0
    assert top_right.x == 500
    assert top_right.y == 100


def test_get_corner_coordinates_width_height_nw(array: np.ndarray):
    origin = Direction.NW
    width = 100
    height = 100

    rectangle = Rectangle(origin=origin, width=width, height=height)
    bottom_left, top_right = rectangle.get_corner_coordinates(array.shape[0], array.shape[1])

    assert bottom_left.x == 0
    assert bottom_left.y == 400
    assert top_right.x == 100
    assert top_right.y == 500


def test_get_corner_coordinates_width_height_ne(array: np.ndarray):
    origin = Direction.NE
    width = 100
    height = 100

    rectangle = Rectangle(origin=origin, width=width, height=height)
    bottom_left, top_right = rectangle.get_corner_coordinates(array.shape[0], array.shape[1])

    assert bottom_left.x == 400
    assert bottom_left.y == 400
    assert top_right.x == 500
    assert top_right.y == 500


def test_get_corner_coordinates_width_height_c(array: np.ndarray):
    origin = Direction.C
    width = 100
    height = 100

    rectangle = Rectangle(origin=origin, width=width, height=height)
    bottom_left, top_right = rectangle.get_corner_coordinates(array.shape[0], array.shape[1])

    assert bottom_left.x == 200
    assert bottom_left.y == 200
    assert top_right.x == 300
    assert top_right.y == 300


def test_get_corner_coordinates_width_height_n(array: np.ndarray):
    origin = Direction.N
    width = 100
    height = 100

    rectangle = Rectangle(origin=origin, width=width, height=height)
    bottom_left, top_right = rectangle.get_corner_coordinates(array.shape[0], array.shape[1])

    assert bottom_left.x == 200
    assert bottom_left.y == 400
    assert top_right.x == 300
    assert top_right.y == 500


def test_get_corner_coordinates_width_height_e(array: np.ndarray):
    origin = Direction.E
    width = 100
    height = 100

    rectangle = Rectangle(origin=origin, width=width, height=height)
    bottom_left, top_right = rectangle.get_corner_coordinates(array.shape[0], array.shape[1])

    assert bottom_left.x == 400
    assert bottom_left.y == 200
    assert top_right.x == 500
    assert top_right.y == 300


def test_get_corner_coordinates_width_height_s(array: np.ndarray):
    origin = Direction.S
    width = 100
    height = 100

    rectangle = Rectangle(origin=origin, width=width, height=height)
    bottom_left, top_right = rectangle.get_corner_coordinates(array.shape[0], array.shape[1])

    assert bottom_left.x == 200
    assert bottom_left.y == 0
    assert top_right.x == 300
    assert top_right.y == 100


def test_get_corner_coordinates_width_height_w(array: np.ndarray):
    origin = Direction.W
    width = 100
    height = 100

    rectangle = Rectangle(origin=origin, width=width, height=height)
    bottom_left, top_right = rectangle.get_corner_coordinates(array.shape[0], array.shape[1])

    assert bottom_left.x == 0
    assert bottom_left.y == 200
    assert top_right.x == 100
    assert top_right.y == 300


def test_get_corner_coordinates_width_sw(array: np.ndarray):
    origin = Direction.SW
    width = 100

    rectangle = Rectangle(origin=origin, width=width)
    bottom_left, top_right = rectangle.get_corner_coordinates(array.shape[0], array.shape[1])

    assert bottom_left.x == 0
    assert bottom_left.y == 0
    assert top_right.x == 100
    assert top_right.y == 500


def test_get_corner_coordinates_height_sw(array: np.ndarray):
    origin = Direction.SW
    height = 100

    rectangle = Rectangle(origin=origin, height=height)
    bottom_left, top_right = rectangle.get_corner_coordinates(array.shape[0], array.shape[1])

    assert bottom_left.x == 0
    assert bottom_left.y == 0
    assert top_right.x == 500
    assert top_right.y == 100


def test_get_corner_coordinates_width_sw_x(array: np.ndarray):
    origin = Direction.SW
    x = 100
    width = 100

    rectangle = Rectangle(origin=origin, width=width, x=x)
    bottom_left, top_right = rectangle.get_corner_coordinates(array.shape[0], array.shape[1])

    assert bottom_left.x == 100
    assert bottom_left.y == 0
    assert top_right.x == 200
    assert top_right.y == 500


def test_get_corner_coordinates_width_sw_y(array: np.ndarray):
    origin = Direction.SW
    y = 100
    width = 100

    rectangle = Rectangle(origin=origin, width=width, y=y)
    bottom_left, top_right = rectangle.get_corner_coordinates(array.shape[0], array.shape[1])

    assert bottom_left.x == 0
    assert bottom_left.y == 100
    assert top_right.x == 100
    assert top_right.y == 500


def test_get_corner_coordinates_width_nw_x(array: np.ndarray):
    origin = Direction.NW
    x = 100
    width = 100

    rectangle = Rectangle(origin=origin, width=width, x=x)
    bottom_left, top_right = rectangle.get_corner_coordinates(array.shape[0], array.shape[1])

    assert bottom_left.x == 100
    assert bottom_left.y == 0
    assert top_right.x == 200
    assert top_right.y == 500


def test_get_corner_coordinates_width_ne_x(array: np.ndarray):
    origin = Direction.NE
    x = -100
    width = 100

    rectangle = Rectangle(origin=origin, width=width, x=x)
    bottom_left, top_right = rectangle.get_corner_coordinates(array.shape[0], array.shape[1])

    assert bottom_left.x == 300
    assert bottom_left.y == 0
    assert top_right.x == 400
    assert top_right.y == 500


def test_get_corner_coordinates_width_se_x_float(array: np.ndarray):
    origin = Direction.SE
    x = -.2
    y = .2
    width = 100

    rectangle = Rectangle(origin=origin, width=width, x=x, y=y)
    bottom_left, top_right = rectangle.get_corner_coordinates(array.shape[0], array.shape[1])

    assert bottom_left.x == 300
    assert bottom_left.y == 100
    assert top_right.x == 400
    assert top_right.y == 500


def test_get_corner_coordinates_width_c_x(array: np.ndarray):
    origin = Direction.C
    x = -100
    width = 100

    rectangle = Rectangle(origin=origin, width=width, x=x)
    bottom_left, top_right = rectangle.get_corner_coordinates(array.shape[0], array.shape[1])

    assert bottom_left.x == 100
    assert bottom_left.y == 0
    assert top_right.x == 200
    assert top_right.y == 500


def test_get_corner_coordinates_width_c_x_float(array: np.ndarray):
    origin = Direction.C
    x = .2
    width = 100

    rectangle = Rectangle(origin=origin, width=width, x=x)
    bottom_left, top_right = rectangle.get_corner_coordinates(array.shape[0], array.shape[1])

    assert bottom_left.x == 300
    assert bottom_left.y == 0
    assert top_right.x == 400
    assert top_right.y == 500


def test_get_corner_coordinates_width_c_y_float(array: np.ndarray):
    origin = Direction.C
    y = .2
    width = 100

    rectangle = Rectangle(origin=origin, width=width, y=y)
    bottom_left, top_right = rectangle.get_corner_coordinates(array.shape[0], array.shape[1])

    assert bottom_left.x == 200
    assert bottom_left.y == 100
    assert top_right.x == 300
    assert top_right.y == 500


def test_get_corner_coordinates_sw_aspect(array: np.ndarray):
    origin = Direction.SW
    aspect = 2

    rectangle = Rectangle(origin=origin, aspect=aspect)
    bottom_left, top_right = rectangle.get_corner_coordinates(array.shape[0], array.shape[1])

    assert bottom_left.x == 0
    assert bottom_left.y == 0
    assert top_right.x == 500
    assert top_right.y == 250


def test_get_corner_coordinates_ne_aspect(array: np.ndarray):
    origin = Direction.NE
    aspect = .5

    rectangle = Rectangle(origin=origin, aspect=aspect)
    bottom_left, top_right = rectangle.get_corner_coordinates(array.shape[0], array.shape[1])

    assert bottom_left.x == 250
    assert bottom_left.y == 0
    assert top_right.x == 500
    assert top_right.y == 500


def test_get_corner_coordinates_height_sw_aspect(array: np.ndarray):
    origin = Direction.SW
    aspect = 2
    height = 200

    rectangle = Rectangle(origin=origin, height=height, aspect=aspect)
    bottom_left, top_right = rectangle.get_corner_coordinates(array.shape[0], array.shape[1])

    assert bottom_left.x == 0
    assert bottom_left.y == 0
    assert top_right.x == 400
    assert top_right.y == 200


def test_get_corner_coordinates_height_c_aspect_y(array: np.ndarray):
    origin = Direction.C
    aspect = .5
    height = 200
    y = 100

    rectangle = Rectangle(origin=origin, height=height, aspect=aspect, y=y)
    bottom_left, top_right = rectangle.get_corner_coordinates(array.shape[0], array.shape[1])

    assert bottom_left.x == 200
    assert bottom_left.y == 250
    assert top_right.x == 300
    assert top_right.y == 450


def test_get_corner_coordinates_width_n_aspect(array: np.ndarray):
    origin = Direction.N
    aspect = 2
    width = 100

    rectangle = Rectangle(origin=origin, width=width, aspect=aspect)
    bottom_left, top_right = rectangle.get_corner_coordinates(array.shape[0], array.shape[1])

    assert bottom_left.x == 200
    assert bottom_left.y == 450
    assert top_right.x == 300
    assert top_right.y == 500


def test_get_corner_coordinates_width_height_e_aspect(array: np.ndarray):
    """
    aspect is ignored with both width and height
    """
    origin = Direction.E
    aspect = 2
    width = 100
    height = 100

    rectangle = Rectangle(origin=origin, width=width, height=height, aspect=aspect)
    bottom_left, top_right = rectangle.get_corner_coordinates(array.shape[0], array.shape[1])

    assert bottom_left.x == 400
    assert bottom_left.y == 200
    assert top_right.x == 500
    assert top_right.y == 300


def test_cook_width_height_e_aspect(array: np.ndarray):
    """
    aspect is ignored with both width and height
    """
    origin = Direction.E
    aspect = 2
    width = 100
    height = 100

    rectangle = Rectangle(origin=origin, width=width, height=height, aspect=aspect)
    cooked_array = rectangle.cook(array)

    assert np.all(cooked_array[400:500, 200:300] == 255)



"""
image and animation processing framework

pyrogis.ingredients -> Ingredient types

pyrogis.kitchen -> cooking "runtime"

pyrogis.restaurant -> feedback/terminal interface handler

pyrogis.course -> animation wrapper
"""



import os

import imageio
import numpy as np
import pytest

from pyrogis.kitchen import Kitchen, Chef


@pytest.fixture
def array():
    return np.asarray([[[200, 200, 200], [30, 30, 30]],
                       [[130, 130, 130], [60, 60, 60]]]).astype(np.dtype('uint8'))


@pytest.fixture
def kitchen(tmp_path) -> Kitchen:
    return Kitchen(Chef, cooked_dir=str(tmp_path / 'cooked'), output_dir=str(tmp_path))


@pytest.fixture
def order_name():
    return 'octo'


@pytest.fixture
def image_path(array: np.ndarray, tmp_path):
    output_path = tmp_path / 'output.png'
    imageio.imwrite(output_path, array)

    return str(output_path)


@pytest.fixture
def animation_path(array, tmp_path) -> str:
    output_path = tmp_path / 'output.mp4'
    imageio.mimwrite(output_path, ims=[array, array])

    return str(output_path)


@pytest.fixture
def dir_path(image_path, tmp_path, order_name) -> str:
    frames_path = tmp_path / 'frames'
    im = imageio.imread(image_path)
    os.mkdir(frames_path)
    imageio.imwrite(frames_path / (order_name + '-1.png'), im=im)
    imageio.imwrite(frames_path / (order_name + '-2.png'), im=im)

    return str(frames_path)


@pytest.fixture
def png_output_path(tmp_path):
    return str(tmp_path / 'output.png')


@pytest.fixture
def gif_output_path(tmp_path):
    return str(tmp_path / 'output.gif')


@pytest.fixture
def mp4_output_path(tmp_path):
    return str(tmp_path / 'output.mp4')



import numpy as np
import pytest

from pyrogis.ingredients.seasonings import Seasoning


@pytest.fixture
def array():
    return np.asarray([[[0, 0, 0], [255, 255, 255]],
                       [[130, 130, 130], [60, 60, 60]]])


def test_cook(array: np.ndarray):
    seasoning = Seasoning()
    cooked_array = seasoning.cook(array)

    assert np.all(cooked_array[0, 1] == 255)



from typing import Dict

import numpy as np
import pytest
from PIL import Image

from pyrogis.ingredients import Dish, Pierogi
from pyrogis.kitchen import Chef, menu
from pyrogis.kitchen.ticket import Ticket, PierogiDesc, IngredientDesc


@pytest.fixture
def pierogi_key():
    return 'pierogi'


@pytest.fixture
def files_key():
    return 'files_key'


@pytest.fixture
def ingredient_key():
    return 'resize'


@pytest.fixture
def pierogi_desc(files_key) -> PierogiDesc:
    return PierogiDesc(files_key=files_key)


@pytest.fixture
def array():
    return np.asarray(
        [
            [[200, 200, 200], [30, 30, 30], [30, 30, 30]],
            [[130, 130, 130], [60, 60, 60], [30, 30, 30]],
            [[200, 200, 200], [30, 30, 30], [30, 30, 30]],
            [[130, 130, 130], [60, 60, 60], [30, 30, 30]]
        ]
    ).astype(np.dtype('uint8'))


@pytest.fixture
def image_path(array: np.ndarray, tmp_path):
    image = Image.fromarray(array)
    output_path = tmp_path / 'output.png'
    image.save(output_path)

    return output_path


@pytest.fixture
def chef():
    return Chef()


@pytest.fixture
def image_paths(files_key, image_path) -> Dict[str, str]:
    files = {
        files_key: image_path
    }
    return files


@pytest.fixture
def ingredient_desc(ingredient_key) -> IngredientDesc:
    return IngredientDesc(ingredient_key, kwargs={})


@pytest.fixture
def pierogi_objects(chef, pierogi_key, pierogi_desc, image_paths):
    pierogi_descs = {
        pierogi_key: pierogi_desc
    }
    return chef.create_pierogi_objects(pierogi_descs, image_paths)


@pytest.fixture
def ticket(pierogi_key, pierogi_desc, ingredient_desc, ingredient_key, image_paths):
    pierogi_descs = {
        pierogi_key: pierogi_desc
    }

    ingredient_descs = {
        ingredient_key: ingredient_desc
    }

    base = pierogi_key
    seasoning_links = {}

    ticket = Ticket(
        pierogis=pierogi_descs,
        files=image_paths,
        ingredients=ingredient_descs,
        base=base,
        recipe=[ingredient_key],
        seasoning_links=seasoning_links
    )

    return ticket


@pytest.fixture
def image_dish(image_path: str) -> Dish:
    return Dish(pierogi=Pierogi.from_path(image_path))


def test_create_pierogi_objects(pierogi_objects):
    assert len(pierogi_objects.values()) == 1


def test_create_ingredient_objects(chef, ingredient_desc, pierogi_objects):
    ingredient_descs = {
        ingredient_key: ingredient_desc
    }
    ingredient_objects = chef.create_ingredient_objects(
        ingredient_descs, pierogi_objects, menu.menu
    )

    assert len(ingredient_objects.values()) == 1


def test_get_ingredient_get(chef, ingredient_desc, ingredient_key):
    ingredients = {
        ingredient_key: menu.menu[ingredient_desc.type_name].type()
    }
    ingredient = chef.get_ingredient(
        ingredients,
        {ingredient_key: ingredient_desc},
        ingredient_key,
        menu
    )

    assert ingredients[ingredient_key] == ingredient


def test_get_ingredient_create(chef, ingredient_desc, ingredient_key):
    ingredients = {}
    ingredient = chef.get_ingredient(
        ingredients,
        {ingredient_key: ingredient_desc},
        ingredient_key,
        menu.menu
    )

    order_type = menu.menu[ingredient_desc.type_name].type

    assert order_type == type(ingredient)


def test_assemble_dish(chef, ticket):
    dish = chef.assemble_ticket(ticket, menu.menu)

    assert dish


def test_cook_dish_image(chef, image_dish):
    dish = chef.cook_dish(image_dish)

    assert dish



import argparse

from .filling import Filling
from .rotate_filling import RotateFilling
from .threshold_filling import ThresholdFilling
from ..ticket import Ticket, IngredientDesc
from ...ingredients import Sort, Ingredient


class SortFilling(Filling):
    type_name: str = 'sort'
    type: Ingredient = Sort

    @classmethod
    def add_parser_arguments(cls, parser: argparse.ArgumentParser):
        """
        add the arguments that a sort parser would need
        """
        RotateFilling.add_parser_arguments(parser)

        ThresholdFilling.add_parser_arguments(parser)

        parser.set_defaults(turns=0)

    @classmethod
    def generate_ticket(
            cls,
            ticket: Ticket,
            path: str = None,
            frame_index: int = 0,
            target_pierogi_uuid: str = None,
            **kwargs
    ):
        """
        add the description of a sort from this path to the dish
        """

        turns = kwargs.pop('turns')
        clockwise = kwargs.pop('clockwise')

        # create rotate description
        rotate_desc = IngredientDesc(
            type_name='rotate',
            kwargs={
                'turns': turns,
                'clockwise': clockwise,
            }
        )
        rotate_uuid = ticket.add_ingredient_desc(rotate_desc)

        # check for implied threshold
        lower_threshold = kwargs.pop('lower_threshold')
        upper_threshold = kwargs.pop('upper_threshold')

        # create threshold desc
        threshold_desc = IngredientDesc(
            type_name='threshold',
            kwargs={
                'lower_threshold': lower_threshold,
                'upper_threshold': upper_threshold,
                'pierogi': target_pierogi_uuid
            }
        )
        # add seasoning link for threshold to season sort
        season_uuid = ticket.add_ingredient_desc(threshold_desc)

        kwargs['rotate'] = rotate_uuid

        ticket = super().generate_ticket(
            ticket, path, frame_index, target_pierogi_uuid, **kwargs
        )

        sort_uuid = ticket.recipe[-1]

        ticket.add_seasoning_link(season_uuid, sort_uuid)

        return ticket



import os
import subprocess
from typing import List

import imageio
import numpy as np

from .ingredients.dish import Dish


class Course:
    """treat a series of Dishes as a unit (animation)"""

    _frames: int = None
    _fps: int = None

    @property
    def frames(self):
        return len(self.dishes)

    def __init__(
            self,
            dishes: List[Dish] = None,
            fps: float = None,
    ):
        self.dishes = dishes
        self.fps = fps

    def serve(self):
        cooked_dishes = []

        for frame in range(self.frames):
            dish = self.dishes[frame]
            # cook with these pixels as first input
            cooked_dish = dish.serve()

            cooked_dishes.append(cooked_dish)

        return Course(cooked_dishes)

    def save(
            self,
            path: str,
            optimize: bool = True,
            duration: float = None,
            fps: float = None
    ) -> None:
        """
        :param duration: ms duration between frames
        """
        if len(self.dishes) > 1:
            if duration is not None:
                fps = 1000 / duration

            if fps is None:
                if self._fps is None:
                    fps = 30
                else:
                    fps = self._fps

            writer = imageio.get_writer(
                path,
                fps=fps
            )

            for dish in self.dishes:
                writer.append_data(np.asarray(dish.pierogi.image))

            writer.close()

            if optimize and os.path.splitext(path)[1] == ".gif":
                try:
                    return_code = subprocess.call(
                        ["gifsicle", '--optimize', path, "--output", path],
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL
                    )
                except:
                    return_code = 1

                if return_code != 0:
                    print("install gifsicle and ensure it's on PATH to optimize gif")

        elif len(self.dishes) == 1:
            self.dishes[0].pierogi.save(path)

        else:
            raise Exception("Dish has no pierogis")



"""
define objects for running the program with rich terminal feedback
"""

import os
import time
from datetime import timedelta
from typing import Callable, Iterable, List

from rich.align import Align
from rich.console import RenderGroup, Console
from rich.live import Live
from rich.panel import Panel
from rich.progress import Progress, BarColumn, TextColumn, Task, ProgressColumn, SpinnerColumn
from rich.rule import Rule
from rich.table import Column
from rich.text import Text
from rich.tree import Tree

from .kitchen.order import Order


class TimeElapsedMsColumn(ProgressColumn):
    """show time elapsed with decimal ms"""

    def __init__(self, decimal_places: int = 1, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.decimal_places = decimal_places

    def render(self, task: Task) -> Text:
        """Show time remaining."""
        elapsed = task.finished_time if task.finished else task.elapsed
        if elapsed is None:
            return Text("-:--:--.---", style="progress.elapsed")
        delta = timedelta(milliseconds=elapsed * 1000)
        split = str(delta).split('.')
        rate_string = "{}.{}".format(split[0], split[1][:self.decimal_places])

        return Text(rate_string, style="progress.elapsed")


class TreeColumn(ProgressColumn):
    """show a tree off of the description with branches"""

    def __init__(self, width: int = 15, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.width = width

    def render(self, task: Task) -> Tree:
        """show tree with order name and input files"""

        tree = Tree(('[bold]' + task.description).center(self.width))

        branches = task.fields.get('branches')

        for path in branches:
            tree.add(os.path.basename(path), style='bold red')

        return tree


class SmoothRateColumn(ProgressColumn):
    """update and display an exponential smoothed rate"""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.completed = 0
        self.smooth_rate = 0
        self.alpha = .2
        self.last_time = None

    def render(self, task: Task) -> Text:
        current_time = time.perf_counter()

        if self.last_time is None:
            self.last_time = time.perf_counter()

        elapsed = current_time - self.last_time

        completed = task.completed
        total = task.total

        # only smooth every second
        if elapsed > 1 and completed != total:
            rate_sample = (completed - self.completed) / elapsed
            self.smooth_rate = self.alpha * rate_sample + (1 - self.alpha) * self.smooth_rate

            self.completed = completed
            self.last_time = current_time

        rate_string = "{rate:>6.2f}🥟/s".format(rate=self.smooth_rate)
        return Text(rate_string)


class Restaurant:
    """run the main program injected with callbacks to display a rich terminal"""

    def __init__(self):
        self.kitchen_tasks = {}
        self.server_tasks = {}

        self.console = Console()

        # for displaying the kitchen's status cooking orders
        self.kitchen_progress = Progress(
            TreeColumn(),
            BarColumn(),
            " [progress.percentage]{task.percentage:>3.0f}% ",
            TimeElapsedMsColumn(),
            SmoothRateColumn()
        )

        # for displaying the stage of orders/kitchen
        self.server_progress = Progress(
            TextColumn(
                "[progress.description]{task.description}",
                justify="center", table_column=Column(width=12),
                style='underline bold'
            ),
            TextColumn("|"),
            SpinnerColumn('arc', speed=1, finished_text='[bold]✓'),
            TextColumn("{task.fields[input_path]}"),
            TextColumn("[bold blue]->"),
            TextColumn("{task.fields[output_path]}")
        )

    def add_kitchen_task(self, order_name: str, input_paths: Iterable[str]):
        task = self.kitchen_progress.add_task(description=order_name, branches=input_paths, cook_rate=0)
        self.kitchen_tasks[order_name] = task
        return task

    def add_server_task(self, order_name: str, input_path: str, output_path: str):
        task = self.server_progress.add_task('...', input_path=input_path, output_path=output_path)
        self.server_tasks[order_name] = task
        return task

    def open(self, run: Callable):
        """open the restaurant using a provided callable"""
        kitchen_group = RenderGroup(Rule("[bold blue]kitchen"),
                                    Align(self.kitchen_progress, align='center'))
        server_group = RenderGroup(Rule("[bold blue]server"),
                                   Align(self.server_progress, align='center'))

        aligned_server_panel = Align(Panel(server_group, expand=False), align='center')
        aligned_kitchen_panel = Align(Panel(kitchen_group), align='center')

        panels = RenderGroup(aligned_kitchen_panel, aligned_server_panel)
        with Live(panels, refresh_per_second=10, console=self.console):
            try:
                run(report_callback=self._report)
            except KeyboardInterrupt:
                pass

    @staticmethod
    def _get_order_name(order: Order) -> str:
        if order.order_name is not None:
            order_name = order.order_name
        else:
            # set order name to first word of input filename if none given
            order_name = os.path.splitext(
                os.path.basename(order.input_path)
            )[0].split()[0]

        return order_name

    def _report(
            self, order: Order, status: str = None,
            completed: int = None, advance: int = None,
            total: int = None, branches: List[str] = None
    ):
        self._update_server(order, status)
        self._update_kitchen(order, completed, total, advance, branches)

    def _update_server(
            self,
            order: Order,
            status: str
    ):
        order_name = self._get_order_name(order)
        server_task = self.server_tasks.get(order_name)

        input_path = os.path.basename(order.input_path)

        if order.output_path is None:
            output_path = '...'
        else:
            output_path = os.path.basename(order.output_path)

        if server_task is None:
            server_task = self.add_server_task(order_name, input_path, output_path)
        if status is not None:
            self.server_progress.update(server_task, description=status)
            self.server_progress.update(server_task, input_path=input_path)
            self.server_progress.update(server_task, output_path=output_path)
            if status == 'done':
                self.server_progress.update(server_task, total=0)

    def _update_kitchen(self, order: Order, completed: int, total: int, advance: int, branches: List[str]):
        order_name = self._get_order_name(order)

        kitchen_task = self.kitchen_tasks.get(order_name)
        if kitchen_task is None:
            kitchen_task = self.add_kitchen_task(order_name, [])
        if completed is not None:
            self.kitchen_progress.update(kitchen_task, completed=completed)
        if total is not None:
            self.kitchen_progress.update(kitchen_task, total=total)
        if advance is not None:
            self.kitchen_progress.update(kitchen_task, advance=advance)
        if branches is not None:
            self.kitchen_progress.update(kitchen_task, branches=branches)



import os

from setuptools import find_packages
from setuptools import setup

readme = open("README.md").read()
changelog = open("CHANGELOG.md").read()

setup_requires = [
    'setuptools-scm>=6.0.1', 'setuptools-rust', 'wheel'
]

if not os.environ.get('READTHEDOCS'):
    from setuptools_rust import RustExtension, Binding

    rust_extensions = [
        RustExtension("pierogis_rs", binding=Binding.PyO3)
    ]
else:
    # don't use RustExtensions
    rust_extensions = None

setup(
    name='pyrogis',
    author="pierogis-live",
    author_email="admin@pierogis.live",
    description="image and animation processing framework",
    url="https://github.com/pierogis/pierogis",
    long_description=readme + "\n\n" + changelog,
    long_description_content_type="text/markdown",
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    install_requires=[
        'Pillow>=8.0.1',
        'numpy>=1.20.1',
        'imageio>=2.9.0',
        'imageio-ffmpeg>=0.4.3',
        'natsort>=7.1.1',
        'rich>=10.1.0'
    ],
    setup_requires=setup_requires,
    extras_require={
        'dev': setup_requires + ['pytest']
    },
    rust_extensions=rust_extensions,
    entry_points={
        'console_scripts': [
            "pyrogis=pyrogis.__main__:main"
        ]
    },
    zip_safe=False
)



import math
import multiprocessing as mp
import os
import signal
import sys
import tempfile
import time
from collections import defaultdict
from typing import Callable, List, Dict

import imageio

from . import menu
from .chef import Cooker
from .menu import Filling
from .order import Order
from .ticket import Ticket
from ..course import Course
from ..ingredients import Pierogi, Dish


def initializer():
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    sys.stdout = open(os.devnull, 'w')


class Kitchen:
    """cook a collection of Ticket objects"""
    menu: Dict[str, Filling] = menu.menu

    _pool: mp.Pool

    def _start_pool(self):
        self.pool = mp.Pool(self.processes, initializer=initializer)

    def __init__(
            self,
            cooker: Cooker,
            processes: int = None,
            cooked_dir: str = 'cooked',
            output_dir: str = '',
            raw_dir: str = None
    ):
        """

        :param cooker:
                Chef-like object to cook with - can be anything that implements the base Chef's methods
        :param processes:
                number of cpus to use (defaults to os.cpu_count)
        :param cooked_dir:
                to output cooked frames
        :param output_dir:
                dir to output fully cooked and plated courses to
        :param raw_dir:
                dir to store raw input files if necessary (defaults to tempfile.mkdtemp)
        """
        self.cooker = cooker
        if processes is None:
            processes = os.cpu_count()
        self.processes = processes
        self.pool = None
        self.cooked_dir = cooked_dir
        self.output_dir = output_dir

        if raw_dir is None:
            raw_dir = tempfile.mkdtemp(suffix='raw')
        self.raw_dir = raw_dir

    def __getstate__(self):
        self_dict = self.__dict__.copy()
        del self_dict['pool']
        return self_dict

    def __setstate__(self, state):
        self.__dict__.update(state)

    @classmethod
    def cook_ticket(
            cls,
            cooker: Cooker,
            ticket: Ticket
    ) -> None:
        """cook a ticket with a thread pool"""
        dish = cooker.assemble_ticket(ticket, cls.menu)
        cooked_dish = cooker.cook_dish(dish)
        cooked_dish.pierogi.save(ticket.output_path)

    def _presave_ticket(self, frame, ticket: Ticket):
        if not os.path.isdir(self.raw_dir):
            os.makedirs(self.raw_dir)

        input_filename = os.path.join(self.raw_dir, os.path.basename(ticket.output_path))
        writer = imageio.get_writer(input_filename)
        writer.append_data(frame)

        pierogi_desc = ticket.pierogis[ticket.base]
        ticket.files[pierogi_desc.files_key] = input_filename
        pierogi_desc.frame_index = 0

    def _auto_pilot(self, order: Order) -> List[Ticket]:
        """test some frames in the animation"""
        tickets = [ticket for ticket in order.tickets if not ticket.skip]

        # test with 5% of the frames, within 2 and 10
        seq_pilot_frames = 2

        frame_index = 0

        next_frame_index = frame_index + seq_pilot_frames
        next_tickets = tickets[frame_index:next_frame_index]
        frame_index = next_frame_index

        # sync cooking
        start = time.perf_counter()
        for ticket in next_tickets:
            self.cook_ticket(self.cooker, ticket)
        seq_rate = seq_pilot_frames / (time.perf_counter() - start)

        if order.presave is None:
            next_frame_index = frame_index + 2
            next_tickets = tickets[frame_index:next_frame_index]
            order.reader.set_image_index(frame_index)
            frame_index = next_frame_index
            # sync cooking with presave frames
            presave_start = time.perf_counter()

            for ticket in next_tickets:
                frame = order.reader.get_next_data()
                self._presave_ticket(frame, ticket)
                self.cook_ticket(self.cooker, ticket)

            presave_rate = seq_pilot_frames / (time.perf_counter() - presave_start)

            if presave_rate > seq_rate:
                order.presave = True
                seq_rate = presave_rate
            else:
                order.presave = False

        if order.cook_async is None:
            par_pilot_frames = max(2, min(round(len(tickets) * .05), 10))

            if order.presave:
                next_frame_index = frame_index + par_pilot_frames
                next_tickets = tickets[frame_index:next_frame_index]
                frame_index = next_frame_index

                results = []

                if self.pool is None:
                    self._start_pool()

                # async cooking with presave frames
                par_presave_start = time.perf_counter()

                for ticket in next_tickets:
                    frame = order.reader.get_next_data()
                    self._presave_ticket(frame, ticket)

                    def error_callback(exception: Exception):
                        order.failures.put((exception, ticket))

                    results.append(self.pool.apply_async(
                        self.cook_ticket, (self.cooker, ticket), error_callback=error_callback
                    ))

                for result in results:
                    result.wait()

                par_rate = par_pilot_frames / (time.perf_counter() - par_presave_start)

            else:
                # parallel cooking

                next_frame_index = frame_index + par_pilot_frames
                next_tickets = tickets[frame_index:next_frame_index]
                frame_index = next_frame_index

                par_start = time.perf_counter()

                results = []

                if self.pool is None:
                    self._start_pool()

                for ticket in next_tickets:
                    def error_callback(exception: Exception):
                        order.failures.put((exception, ticket))

                    results.append(self.pool.apply_async(
                        self.cook_ticket, (self.cooker, ticket), error_callback=error_callback
                    ))

                for result in results:
                    result.wait()

                par_rate = par_pilot_frames / (time.perf_counter() - par_start)

            if par_rate > seq_rate:
                order.cook_async = True

        next_tickets = tickets[frame_index:]

        return next_tickets

    def _set_output_paths(self, order: Order):
        frames = len(order.tickets)

        digits = math.floor(math.log(frames, 10)) + 1

        cooked_dir = self.cooked_dir

        if not os.path.isdir(cooked_dir):
            os.makedirs(cooked_dir)

        suborders = defaultdict(list)

        for ticket in order.tickets:
            suborder_name = ticket.input_filename

            suborders[suborder_name].append(ticket)

        for suborder_name, suborder in suborders.items():
            suborder_length = len(suborder)

            frame_index = 0

            for ticket in suborder:
                if suborder_length > 1:
                    padded_frame_index = str(frame_index + 1).zfill(digits)
                    frame_suffix = '-' + padded_frame_index
                else:
                    frame_suffix = ''

                output_path = os.path.join(
                    cooked_dir,
                    '{suborder_base}{frame_suffix}{extension}'.format(
                        suborder_base=os.path.splitext(suborder_name)[0],
                        frame_suffix=frame_suffix,
                        extension='.png'
                    )
                )

                if os.path.isfile(output_path):
                    input_is_output = os.path.samefile(ticket.input_path, output_path)

                    # this is really rough and can be solved by a system
                    # for manipulating courses/animations as a whole
                    if order.resume:
                        ticket.skip = True
                    elif input_is_output:
                        if not os.path.isdir(self.raw_dir):
                            os.makedirs(self.raw_dir)

                        raw_input_path = os.path.join(
                            self.raw_dir, os.path.basename(ticket.input_path)
                        )
                        os.rename(ticket.input_path, raw_input_path)
                        ticket.input_path = raw_input_path
                    else:
                        os.remove(output_path)

                ticket.output_path = output_path
                frame_index += 1

    def queue_order(
            self, order: Order, start_callback: Callable, report_status: Callable
    ):
        self._set_output_paths(order)
        self.processes = order.processes

        start_callback()

        report_status(order, status='preprocessing')

        if not os.path.isfile(order.input_path):
            order.presave = False

        if len(order.tickets) > 8 and (order.presave is None or order.cook_async is None):
            next_tickets = self._auto_pilot(order)
        else:
            next_tickets = order.tickets

        report_status(order, status='cooking')

        for ticket in next_tickets:
            if order.presave:
                frame = order.reader.get_next_data()
                self._presave_ticket(frame, ticket)

            if order.cook_async:
                if self.pool is None:
                    self._start_pool()

                def error_callback(exception: Exception):
                    order.failures.put((exception, ticket))

                self.pool.apply_async(
                    func=self.cook_ticket,
                    args=(self.cooker, ticket),
                    error_callback=error_callback
                )

            else:
                self.cook_ticket(self.cooker, ticket)

        if self.pool is not None:
            self.pool.close()
            self.pool.join()

    def plate(
            self,
            order: Order
    ) -> str:
        """"""
        dishes = []

        if len(order.tickets) == 0:
            raise Exception("Order has no tickets")

        for ticket in order.tickets:
            frame_path = ticket.output_path

            dish = Dish(pierogi=Pierogi.from_path(path=frame_path))

            dishes.append(dish)

        course = Course(dishes=dishes)

        fps = order.fps
        optimize = order.optimize
        frame_duration = order.duration

        output_path = os.path.join(self.output_dir, os.path.basename(order.output_path))

        dup_index = 1

        dup_trial_path = output_path
        base, ext = os.path.splitext(output_path)

        while os.path.isfile(dup_trial_path):
            dup_trial_path = base + '-' + str(dup_index) + ext

            dup_index += 1

        order.output_path = dup_trial_path

        course.save(
            order.output_path,
            optimize,
            duration=frame_duration,
            fps=fps
        )

        return output_path



import os
from multiprocessing import Queue
from pathlib import Path
from typing import List, Union, Tuple

import imageio

from .ticket import Ticket


class Order:
    order_name: str
    tickets: List[Ticket]
    failures: "Queue[Tuple[Exception, Ticket]]"
    output_path: str
    fps: float
    resume: bool = False
    presave: bool = None
    cook_async: bool = None
    processes: int = None
    _reader = None

    @property
    def order_name(self):
        if self._order_name is None:
            # set order name to first word of input filename if none given
            if os.path.isfile(self.input_path):
                self._order_name = os.path.splitext(
                    os.path.basename(self.input_path)
                )[0].split()[0]

        return self._order_name

    @property
    def reader(self):
        if self._reader is None:
            if os.path.isfile(self.input_path):
                self._reader = imageio.get_reader(self.input_path)

        return self._reader

    @property
    def ticket_output_paths(self):
        for ticket in self.tickets:
            yield ticket.output_path

    @property
    def frames(self) -> int:
        return len(self.tickets)

    @property
    def output_path(self) -> str:
        order_name = self.order_name

        if self._output_path is None:
            if order_name is None:
                order_name = os.path.splitext(os.path.basename(self.input_path))[0]
            if self.frames == 1:
                self._output_path = order_name + '.png'
            elif self.frames == 0:
                pass
            else:
                self._output_path = order_name + '.gif'

        if self._output_path is not None:
            self._output_path = os.path.expanduser(
                os.path.abspath(self._output_path)
            )
        return self._output_path

    @output_path.setter
    def output_path(self, value: str) -> None:
        self._output_path = value

    def __init__(
            self,
            order_name: str,
            input_path: str,
            output_path: Union[str, Path] = None,
            fps: float = None,
            duration: int = None,
            optimize: bool = None,
            presave: bool = None,
            cook_async: bool = None,
            processes: int = None,
            resume: bool = None,
    ):
        self._order_name = order_name
        self.input_path = input_path
        self.tickets = []
        self.failures = Queue()
        self._output_path = output_path
        self.fps = fps
        self.duration = duration
        self.optimize = optimize
        self.presave = presave
        self.cook_async = cook_async
        self.processes = processes
        self.resume = resume

    def add_ticket(self, ticket: Ticket):
        self.tickets.append(ticket)



import os

import pytest

from pyrogis.kitchen import Order, Kitchen, Ticket


@pytest.fixture
def order(order_name: str, dir_path: str):
    order = Order(
        order_name,
        'input.mp4',
        fps=25,
        optimize=True
    )

    for filename in os.listdir(dir_path):
        if filename.startswith(order_name):
            order.add_ticket(
                Ticket(output_path=os.path.join(dir_path, filename)),
            )

    return order


def test_plate_gif(kitchen: Kitchen, order: Order, gif_output_path: str):
    order.output_name = gif_output_path

    output_path = kitchen.plate(
        order
    )

    assert os.path.isfile(output_path)


def test_plate_mp4(kitchen: Kitchen, order: Order, mp4_output_path: str):
    order.output_path = mp4_output_path

    output_path = kitchen.plate(
        order
    )

    assert os.path.isfile(output_path)
#
#
# def test_plate_dir(kitchen: Kitchen, resource_dir: str, tmp_path):
#     input_path = os.path.join(resource_dir, 'frames')
#     output_path = tmp_path / 'output.mp4'
#     optimize = True
#
#     order = Order(
#         'octo',
#         input_path,
#         output_path=output_path,
#         fps=25,
#         optimize=optimize
#     )
#
#     order.tickets = [
#         Ticket(output_path=os.path.join(input_path, 'octo-01.png')),
#         Ticket(output_path=os.path.join(input_path, 'octo-02.png'))
#     ]
#
#     kitchen.plate(order)
#
#     assert os.path.isfile(output_path)



import numpy as np
import pytest

from pyrogis.ingredients import Dish, Pierogi, Recipe


@pytest.fixture
def array():
    return np.asarray([[[200, 200, 200], [30, 30, 30]],
                       [[130, 130, 130], [60, 60, 60]]]).astype(np.dtype('uint8'))


@pytest.fixture
def image_file():
    return 'resources/gnome.jpg'


@pytest.fixture
def animation_file():
    return 'resources/octo.mp4'


def test_dish_pierogis_single(array):
    pierogi = Pierogi(pixels=array)

    dish = Dish(pierogi=pierogi)

    assert np.all(dish.pierogi.pixels == pierogi.pixels)


def test_dish_serve(array):
    pierogi = Pierogi(pixels=array)

    dish = Dish(pierogi=pierogi, recipe=Recipe())
    cooked_dish = dish.serve()

    assert np.all(cooked_dish.pierogi.pixels == pierogi.pixels)



from .custom_filling import CustomFilling
from .filling import Filling
from .quantize_filling import QuantizeFilling
from .resize_filling import ResizeFilling
from .rotate_filling import RotateFilling
from .sort_filling import SortFilling
from .threshold_filling import ThresholdFilling
from .crop_filling import CropFilling

menu = {
    SortFilling.type_name: SortFilling,
    QuantizeFilling.type_name: QuantizeFilling,
    'custom': CustomFilling,
    ThresholdFilling.type_name: ThresholdFilling,
    ResizeFilling.type_name: ResizeFilling,
    'rotate': RotateFilling,
    'crop': CropFilling
}



import math
from typing import Union, Tuple

import numpy as np

from .cartography import Direction, Coordinate
from .seasoning import Seasoning


class Rectangle(Seasoning):
    width: Union[int, float]
    height: Union[int, float]
    x: Union[int, float]
    y: Union[int, float]
    aspect: float
    origin: Direction

    ORIGIN: Direction = Direction.SW

    def prep(
            self,
            width: Union[int, float] = None, height: Union[int, float] = None,
            x: Union[int, float] = 0, y: Union[int, float] = 0,
            aspect: float = None, origin: Direction = ORIGIN,
            **kwargs
    ):
        """
        :param x: 0 = left
        :param y: 0 = bottom
        :param width: width in pixels
        :param height: height in pixels
        :param aspect: aspect ratio to fill missing height or width (width/height, 1 means square)
        :param origin: location of origin for direction of crop select
        """
        super().prep(**kwargs)

        self.width = width
        self.height = height
        self.x = x
        self.y = y
        self.aspect = aspect
        self.origin = origin

    def cook(self, pixels: np.ndarray):
        # create an array of all excludes

        boolean_array = np.zeros(pixels.shape[:2], dtype=bool)

        bottom_left, top_right = self.get_corner_coordinates(pixels.shape[0], pixels.shape[1])

        boolean_array[bottom_left.x:top_right.x, bottom_left.y:top_right.y] = True

        boolean_array = np.expand_dims(boolean_array, 2)

        binary_pixels = np.where(
            boolean_array, self.include_pixel, self.exclude_pixel
        )

        return binary_pixels

        # include all pixels that are inside of the bounding coords

    def get_corner_coordinates(
            self, bounding_width: int, bounding_height: int
    ) -> Tuple[Coordinate, Coordinate]:
        # x, y for bottom left and top right

        if self.origin is Direction.SW:
            x = 0
            y = 0
            bl_x_multiplier = 0
            tr_x_multiplier = 1
            bl_y_multiplier = 0
            tr_y_multiplier = 1

        elif self.origin is Direction.SE:
            x = bounding_width
            y = 0
            bl_x_multiplier = 1
            tr_x_multiplier = 0
            bl_y_multiplier = 0
            tr_y_multiplier = 1

        elif self.origin is Direction.NW:
            x = 0
            y = bounding_height
            bl_x_multiplier = 0
            tr_x_multiplier = 1
            bl_y_multiplier = 1
            tr_y_multiplier = 0

        elif self.origin is Direction.NE:
            x = bounding_width
            y = bounding_height
            bl_x_multiplier = 1
            tr_x_multiplier = 0
            bl_y_multiplier = 1
            tr_y_multiplier = 0

        elif self.origin is Direction.C:
            x = bounding_width * .5
            y = bounding_height * .5
            bl_x_multiplier = .5
            tr_x_multiplier = .5
            bl_y_multiplier = .5
            tr_y_multiplier = .5

        elif self.origin is Direction.N:
            x = bounding_width * .5
            y = bounding_height
            bl_x_multiplier = .5
            tr_x_multiplier = .5
            bl_y_multiplier = 1
            tr_y_multiplier = 0

        elif self.origin is Direction.E:
            x = bounding_width
            y = bounding_height * .5
            bl_x_multiplier = 1
            tr_x_multiplier = 0
            bl_y_multiplier = .5
            tr_y_multiplier = .5

        elif self.origin is Direction.S:
            x = bounding_width * .5
            y = 0
            bl_x_multiplier = .5
            tr_x_multiplier = .5
            bl_y_multiplier = 0
            tr_y_multiplier = 1

        elif self.origin is Direction.W:
            x = 0
            y = bounding_height * .5
            bl_x_multiplier = 0
            tr_x_multiplier = 1
            bl_y_multiplier = .5
            tr_y_multiplier = .5

        else:
            raise Exception()

        if self.x != 0:
            if 1 > self.x > -1:
                x += self.x * bounding_width
            else:
                x += self.x

        if self.y != 0:
            if 1 > self.y > -1:
                y += self.y * bounding_height
            else:
                y += self.y

        origin = Coordinate(x, y)

        width = None
        height = None

        if self.width is not None:
            width = self.width
            if 1 > width > -1:
                width *= bounding_width

        if self.height is not None:
            height = self.height
            if 1 > height > -1:
                height *= bounding_height

        if not (self.width is not None and self.height is not None):
            if self.width is not None:
                if self.aspect is not None:
                    height = width / self.aspect
                else:
                    height = bounding_height
            elif self.height is not None:
                if self.aspect is not None:
                    width = height * self.aspect
                else:
                    width = bounding_width

            else:
                width = bounding_width
                height = bounding_height

                if self.aspect is not None:
                    if width / height > self.aspect:
                        width = height * self.aspect
                    else:
                        height = width / self.aspect

        bl_x = int(math.ceil(origin.x - bl_x_multiplier * width))
        if bl_x < 0:
            bl_x = 0

        bl_y = int(math.ceil(origin.y - bl_y_multiplier * height))
        if bl_y < 0:
            bl_y = 0

        tr_x = int(math.ceil(origin.x + tr_x_multiplier * width))
        if tr_x > bounding_width:
            tr_x = bounding_width

        tr_y = int(math.ceil(origin.y + tr_y_multiplier * height))
        if tr_y > bounding_height:
            tr_y = bounding_height

        bottom_left = Coordinate(
            bl_x,
            bl_y
        )
        top_right = Coordinate(
            tr_x,
            tr_y
        )

        return bottom_left, top_right



from .filling import Filling
from ...ingredients import Rotate


class RotateFilling(Filling):
    type_name = 'rotate'
    type = Rotate

    @classmethod
    def add_parser_arguments(cls, parser):
        """
        add palette and palette size to this parser
        add rscolorq params to the parser and parent
        """

        # add palette and palette size
        parser.add_argument('-t', '--turns', default=1, type=int)
        parser.add_argument(
            '--ccw',
            dest='clockwise',
            action='store_false'
        )



import argparse
from abc import ABC, abstractmethod

from ..ticket import Ticket, IngredientDesc
from ...ingredients import Ingredient


class Filling(ABC):
    """bundles a parser with a method for generating a ticket"""
    type_name: str = None
    type: Ingredient = None

    @classmethod
    def generate_ticket(
            cls,
            ticket: Ticket,
            path: str = None,
            frame_index: int = 0,
            target_pierogi_uuid: str = None,
            **kwargs
    ) -> Ticket:
        """
        add a description of a quantize recipe
        """

        if path is not None:
            target_pierogi_uuid = ticket.add_pierogi(path, frame_index)
            ticket.base = target_pierogi_uuid

        if cls.type_name is not None:
            ingredient_desc = IngredientDesc(
                type_name=cls.type_name,
                kwargs=kwargs
            )

            ingredient_uuid = ticket.add_ingredient_desc(ingredient_desc)

            ticket.extend_recipe([ingredient_uuid])

        return ticket

    @classmethod
    def get_parser(cls):
        """
        get a parser for this menu item
        """
        parser = argparse.ArgumentParser(add_help=False)
        parser.set_defaults(generate_ticket=cls.generate_ticket)

        parser.add_argument(
            '--presave',
            action='store_true',
            default=None,
            help="presave frames for cooking animations"
        )
        parser.add_argument(
            '--async',
            action='store_true',
            default=None,
            help="use multiple python processes to cook frames"
        )
        parser.add_argument(
            '--processes',
            type=int,
            help="number of async processes to use"
        )
        parser.add_argument(
            '--resume',
            action='store_true',
            help="resume from already cooked frames in the cooked directory with the same order name"
        )

        # get extra parser arguments from subclasses
        cls.add_parser_arguments(parser)

        return parser

    @classmethod
    @abstractmethod
    def add_parser_arguments(cls, parser):
        pass
